-module(icfp_print).

-include("icfp.hrl").

-define(MAX_RANK_LOOKAHEAD, 10).	% pretty arbitrary

-import(lists, [map/2, foldl/3, reverse/1]).
-export([file/3]).

%% Stringify the parse tree (Parse) from initial state Istate and
%% write the result to Outfile.
file(Outfile, Parse, Istate) ->
    file:write_file(Outfile, print(Parse, Istate)).

print(Parse, State) ->
    io:format("Parse has ~p nodes.~n", [length(Parse)]),
    print(Parse, State, []).

%% Stack is a list of tags, like 'b', 5, 'B', 'TT', etc
print([], S, Stack) ->
    %% end of input; unwind the stack to be finished
    [format_pop(Tag) || Tag <- Stack];
print([{SNew, Text}|T], SOld, Stack0) ->
    Ops = ops(SOld, SNew, Stack0),
    {Pushes, Pops, Stack} = process_ops(Ops, Stack0, SNew, T),
    PopText = map(fun(Tag) -> format_pop(Tag) end, Pops),
    PushText = map(fun(Tag) -> format_push(Tag) end, Pushes),
    [PopText, PushText, Text | print(T, SNew, reverse(Pushes) ++ Stack)].

%% Returns: {Pushes, Pops, Stack}
process_ops(Ops, Stack, SNew, Parse) ->
    Pops0 = [Tag || {pop, Tag} <- Ops],
    {Pops1, Repushes, Stack1} = process_pops(Pops0, Stack),
    Pushes0 = [Tag || {push, Tag} <- Ops],
    Pushes1 = filter_pushes(Repushes ++ Pushes0),
    Pushes2 = rank_sort_pushes(Pushes1, Parse, SNew),
    {Pushes2, Pops1, Stack1}.

%% process_pops(PopTags, Stack) => {SortedPops, Repushes, Stack'}
%% `Repushes' are elements that were lifted to remove others beneath
%% them - they need to be pushed back on.
process_pops(Pops, Stack) ->
    process_pops(Pops, Stack, [], []).

process_pops([], Stack, Acc, Repushes) ->
    {reverse(Acc), Repushes, Stack};
process_pops(Pops, [H|T], Acc, Repushes) ->
    case lists:member(H, Pops) of
	true  -> process_pops(Pops -- [H], T, [H|Acc], Repushes);
	false -> process_pops(Pops, T, [H|Acc], [H|Repushes])
    end.

%% Returns: Pushes'
%% Try to order Pushes so that they'll be popped off in the right order.
rank_sort_pushes(Pushes, Parse, S0) ->
    Ranked = rank_pushes(Pushes, Parse, S0, 1),
    Sorted = reverse(lists:keysort(1, Ranked)),
    map(fun({Rank, Push}) -> Push end,
	Sorted).

rank_pushes(Pushes, Parse, _, Rank) when Parse == [];
					 Rank == ?MAX_RANK_LOOKAHEAD ->
    map(fun(Push) -> {?MAX_RANK_LOOKAHEAD, Push} end,
	Pushes);
rank_pushes(Pushes, [{S,_}|T], S0, Rank) ->
    F = fun(Push, {Stops, Gos}) ->
		case stops_in(Push, S0, S) of
		    true ->
			{[Push|Stops], Gos};
		    false ->
			{Stops, [Push|Gos]}
		end
	end,
    {Stops, Gos} = foldl(F, {[],[]}, Pushes),
    map(fun(Stop) -> {Rank,Stop} end, Stops) ++ rank_pushes(Gos,T,S0,Rank+1);
rank_pushes([], _, _, _) ->
    [].

%% There is some stuff that's been accidentally pushed off the
%% stack. Instead of pushing it all off, we take a look to see if
%% there's anything best left off - if we have multiple colours or
%% font sizes, we can just keep the most current one.
filter_pushes(Repushes) ->
    filter_pushes(Repushes, [], []).

filter_pushes([], ColorL, SizeL) ->
    ColorL ++ SizeL;
filter_pushes([H|T], ColorL, SizeL) ->
    case is_color(H) of
	true ->
	    filter_pushes(T, [H], SizeL);
	false ->
	    if integer(H) ->
		    filter_pushes(T, ColorL, [H]);
	       true ->
		    [H|filter_pushes(T, ColorL, SizeL)]
	    end
    end.

%% stops_in(Tag, TagState, NewState) => true | false
%% See if the Tag stops in NewState (i.e. must be popped to get there)
stops_in('B', _, #s{b=true})  -> false;
stops_in('B', _, #s{b=false}) -> true;
stops_in('EM', _, #s{em=false, s=false}) -> true;
stops_in('EM', _, _)                     -> false;
stops_in('I', _, #s{i=true})  -> false;
stops_in('I', _, #s{i=false}) -> true;
stops_in('S', _, #s{s=true})  -> false;
stops_in('S', _, #s{s=false}) -> true;
stops_in('TT', _, #s{tt=true})  -> false;
stops_in('TT', _, #s{tt=false}) -> true;
stops_in('U', #s{u=Old}, #s{u=New}) -> New < Old;
stops_in(N, _, #s{size=N}) when integer(N) -> false;
stops_in(N, _, _) when integer(N)          -> true;
stops_in(Color, _, #s{color=Color}) -> false;
stops_in(Color, _, _)                -> true.

%% ops(OldState, NewState, Stack) => [Op]
%% Decides which push and pop operations are needed to go from state
%% Old to state New. Returns them in a list of {push, Tag} and {pop,
%% Tag} tuples.
ops(Old, New, Stack) ->
    lists:flatmap(fun(Item) -> tag(Item, Stack, Old, New) end,
		  record_info(fields, s)).

%% tag(Field, OldState, NewState) => [Op]
tag(b, _, #s{b=X}, #s{b=Y}) ->
    change_bool('B', X, Y);
tag(em, _, #s{em=X}, #s{em=Y}) ->
    change_bool('EM', X, Y);
tag(i, _, #s{i=X}, #s{i=Y}) ->
    change_bool('I', X, Y);
tag(s, _, #s{s=X}, #s{s=Y}) ->
    change_bool('S', X, Y);
tag(tt, _, #s{tt=X}, #s{tt=Y}) ->
    change_bool('TT', X, Y);
tag(u, _, #s{u=X}, #s{u=Y}) ->
    if
	X == Y -> [];
	X > Y  -> lists:duplicate(X - Y, {pop, 'U'}); % want less underline
	X < Y  -> lists:duplicate(Y - X, {push, 'U'}) % want more underline
    end;
tag(size, Stack, #s{size=X}, #s{size=Y}) ->
    if
	X == Y         -> [];
	X == undefined -> [{push, Y}];
	Y == undefined -> [{pop, I} || I <- Stack, integer(I)];
	length(Stack) < 2 ->
	    [{push, Y}];
	true->
	    H = hd(Stack),
	    if
		integer(H) ->
		    case first_size(tl(Stack)) of
			Y -> [{pop, H}];
			_ -> [{push, Y}]
		    end;
		true ->
		    [{push, Y}]
	    end
    end;
tag(color, Stack, S0 = #s{color=X}, S1 = #s{color=Y}) ->
    if
	X == Y         -> [];
	X == undefined -> [{push, Y}];
	Y == undefined -> [{pop, C} || C <- Stack, is_color(C)];
	length(Stack) < 2 ->
	    [{push, Y}];
	true ->
	    H = hd(Stack),
	    case is_color(H) of
		true ->
		    case first_color(tl(Stack)) of
			Y ->
			    %% if the top of the stack is another color, and
			    %% underneath it is the color we want, then just
			    %% pop
			    [{pop, H}];
			_ ->
			    %% Otherwise push the color
			    [{push, Y}]
		    end;
		_ ->
		    [{push, Y}]
	    end
    end.

first_size([])                    -> not_found;
first_size([H|T]) when integer(H) -> H;
first_size([_|T])                 -> first_size(T).

first_color([]) -> not_found;
first_color([H|T]) ->
    case is_color(H) of
	true  -> H;
	false -> first_color(T)
    end.

is_color(C) -> lists:member(C, [r,g,b,c,m,y,k,w]).

%% change_bool(Tag, Old, New)
%% Determine the Ops for setting a boolean value of Tag from Old to New
change_bool(Tag, X, X)        -> [];
change_bool(Tag, true, false) -> [{pop, Tag}];
change_bool(Tag, false, true) -> [{push, Tag}].

%% Text formatting
format_pop(Tag) ->
    ["</", tag_to_string(Tag), ">"].

format_push(Tag) ->
    ["<", tag_to_string(Tag), ">"].

tag_to_string(I) when integer(I) ->
    integer_to_list(I);
tag_to_string(A) when atom(A) ->
    atom_to_list(A).

